#ifndef LIBcDAG_HPP
#define LIBcDAG_HPP

#include <mpi.h>

#ifdef HAVE_PGT
#include <libpgt.h>
#endif

extern int G_cDAG_Initialized;
extern int G_cDAG_MyRank;
extern int G_cDAG_GroupSize;
extern int G_cDAG_Tag;
extern MPI_Comm G_cDAG_WorldComm;

#include "cDAG_Types.hpp"
#include "cDAG_Graph.hpp"

// if neither TRAN_MPI nor TRAN_OF is defined
// TRAN_MPI should be selected as default

#ifndef TRAN_MPI
#ifndef TRAN_OF
#define TRAN_MPI
#endif
#endif

#ifdef TRAN_OF
#include "cDAG_Tran_OFED.hpp"
#endif
#ifdef TRAN_MPI
#include "cDAG_Tran_MPI.hpp"
#endif

typedef int cDAG_Vertex;

cDAG_Graph cDAG_Create_graph();
cDAG_Graph _cDAG_Create_graph_internal();
void cDAG_Requires(cDAG_Graph g, cDAG_Vertex a, cDAG_Vertex b);
cDAG_Schedule cDAG_Compile(cDAG_Graph g); 
cDAG_Schedule cDAG_Compile_internal(cDAG_Graph g); 
cDAG_Handle cDAG_Run(cDAG_Schedule sched);

int cDAG_FreeGraph(cDAG_Graph g);

int cDAG_Test(cDAG_Handle handle);
int cDAG_Wait(cDAG_Handle handle);

void cDAG_ReuseSchedule(cDAG_Schedule sched);

static inline int cDAG_Init() {

  MPI_Comm comm=MPI_COMM_WORLD;

  G_cDAG_Initialized=1;
  MPI_Comm_rank(comm,&G_cDAG_MyRank);
  MPI_Comm_size(comm,&G_cDAG_GroupSize);

  MPI_Comm_dup(MPI_COMM_WORLD, &G_cDAG_WorldComm);

  char pgtname[128];
  snprintf(pgtname, 127, "goal_%i_perf", G_cDAG_GroupSize);

#ifdef HAVE_PGT
  PGT_Init(pgtname, 0);
#endif

  return cDAG_SUCCESS;
}

static inline int cDAG_IsInitialized() {
  return G_cDAG_Initialized;
}

static inline int cDAG_GroupSize() {
  return G_cDAG_GroupSize;
}

static inline int cDAG_MyRank() {
  return G_cDAG_MyRank;
}

static inline int cDAG_Finalize() {
  MPI_Comm_free(&G_cDAG_WorldComm);
  G_cDAG_Initialized=-2;

#ifdef HAVE_PGT
  PGT_Finalize();
#endif
  
  return 0;
};

static inline int cDAG_IsFinalized() {
  if(G_cDAG_Initialized == -2) return 1;
  else return 0;
}

static inline int cDAG_Send(cDAG_Graph graph, void* buf, int count, int dest, cDAG_MemType mem=cDAG_USERSPACE) {

  /* store the passed arguments */
  NBC_Args_send send_args;
  send_args.buf=buf;
  send_args.memtype=mem;
  send_args.count=count;
  send_args.dest=dest;

  return add_vertex(graph, T_SEND, &send_args);
}

static inline int cDAG_Recv(cDAG_Graph graph, void* buf, int count, int source, cDAG_MemType mem=cDAG_USERSPACE) {

  /* store the passed arguments */
  NBC_Args_recv recv_args;
  recv_args.buf=buf;
  recv_args.memtype=mem;
  recv_args.count=count;
  recv_args.source=source;

  return add_vertex(graph, T_RECV, &recv_args);
}

static inline int cDAG_RegisterOp(cDAG_Graph graph, void (*func)(void*, void*), int argsize) {

	graph->num_ops++;
	if (graph->num_ops > graph->alloc_ops) {
		graph->alloc_ops *= 2;
		graph->ops = (struct op*) realloc(graph->ops, graph->alloc_ops * sizeof(struct op));
	}
	int ind = graph->num_ops - 1;
	graph->ops[ind].func = func;
	graph->ops[ind].argsize = argsize;

	return ind;

}

static inline int cDAG_Op(cDAG_Graph graph, cDAG_Opcode op, void* myargs) {

	NBC_Args_op op_args;

	if (op >= 0) {
		op_args.opnum = op;
		op_args.func = graph->ops[op].func;
		op_args.argsize = graph->ops[op].argsize;
		op_args.args = malloc(graph->ops[op].argsize);
		memcpy(op_args.args, myargs, graph->ops[op].argsize);
	}
	else {
		op_args.opnum = op;
		op_args.func = NULL;
		op_args.argsize = predefined_op_get_argsize(op);
		op_args.args = malloc(predefined_op_get_argsize(op));
		memcpy(op_args.args, myargs, predefined_op_get_argsize(op));
	}

	return add_vertex(graph, T_OP, &op_args);
}

static inline void cDAG_AllocateScratchpad(cDAG_Graph graph, unsigned int size) {
	graph->scratchpad_size = size;
}

static inline void cDAG_FreeSchedule(cDAG_Schedule sched) {
	free(sched.active);
	free(sched.copy);
}

#endif
